/**CFile***********************************************************************

  FileName    [bmcRbc.c]

  PackageName [bmc]

  Synopsis    [Implementation for generic interface to rbc. Temporary located
  in the <tt>bmc</tt> package]

  Description [This implementation depends on the rbc package is being used. 
  Implementation is currently a thin wrapper around rbc supplied interface, so
  functions are really minimal]

  SeeAlso     []

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"



/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
#define RBC_SPEC_TYPE_P RBC_SPEC_TYPE*
#define RBC_SPEC_MANAGER_TYPE_P RBC_SPEC_MANAGER_TYPE*

typedef RBC_SPEC_BOOL_TYPE rbc_bool_t; /* this type is not exported */ 



/**Struct**********************************************************************

  Synopsis    [shift_memoize_key is a provate struct used in shift memoizing]

  Description [Used in order to contain the key values in hash operation 
  involved in shift memoizing]

  SeeAlso     []

******************************************************************************/
typedef struct shift_memoize_key_TAG
{
  rbc_ptr rbc;
  int     shift;
} shift_memoize_key; 


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
static st_table* GL_phtRbcShift = NULL; /* shift hash table */


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Declarations of internal functions                                        */
/*---------------------------------------------------------------------------*/

static rbc_ptr rbc_int_makeAnd ARGS((RbcManager_ptr rbcManager, rbc_ptr left, 
				     rbc_ptr right, rbc_bool_t sign));
static rbc_ptr rbc_int_makeOr ARGS((RbcManager_ptr rbcManager, rbc_ptr left, 
				    rbc_ptr right, rbc_bool_t sign));
static rbc_ptr rbc_int_makeIff ARGS((RbcManager_ptr rbcManager, rbc_ptr left, 
				     rbc_ptr right, rbc_bool_t sign));
static rbc_ptr rbc_int_makeXor ARGS((RbcManager_ptr rbcManager, rbc_ptr left, 
				     rbc_ptr right, rbc_bool_t sign));
static rbc_ptr rbc_int_makeIte ARGS((RbcManager_ptr rbcManager, rbc_ptr c, 
				     rbc_ptr t, rbc_ptr e, rbc_bool_t sign));

static rbc_ptr rbc_getLeftOpnd ARGS((rbc_ptr f));
static rbc_ptr rbc_getRightOpnd ARGS((rbc_ptr f));

static void rbc_mark ARGS((RbcManager_ptr rbc, rbc_ptr f));
static void rbc_unmark ARGS((RbcManager_ptr rbc, rbc_ptr f));


/* Private structures gateway: following functions convert the 
   implementation-dependent rbc structures to the interface-owned 
   generic ones, and viceversa as possible */
static rbc_ptr  convert_to_generic_rbc ARGS((RBC_SPEC_TYPE_P pRbc));

static RBC_SPEC_TYPE_P convert_to_specific_rbc ARGS((rbc_ptr pRbc));

static rbc_ptr*  convert_to_generic_rbc_ref ARGS((RBC_SPEC_TYPE_P* pRbcRef));

static RBC_SPEC_TYPE_P* convert_to_specific_rbc_ref ARGS((rbc_ptr* pRbcRef));

static RBC_SPEC_MANAGER_TYPE_P 
convert_to_specific_rbc_manager ARGS((RbcManager_ptr pRbcManager)); 

static RbcManager_ptr 
convert_to_generic_rbc_manager ARGS((RBC_SPEC_MANAGER_TYPE_P pRbcManager)); 

static RBC_SPEC_BOOL_TYPE 
convert_to_specific_rbc_bool ARGS((rbc_bool_t rbcBool));

static rbc_bool_t  
convert_to_generic_rbc_bool  ARGS((RBC_SPEC_BOOL_TYPE rbcBool));


/* Hash table for memoizing: internal function */
static int 
st_shift_memoize_key_cmp ARGS((const char* _key1, const char* _key2));

static int 
st_shift_memoize_key_hash ARGS((char* _key, const int size));

static enum st_retval 
bmcShiftHash_CallbackDeleteEntryAndKey ARGS((char* key, char* record, 
					     char* dummy));

 

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

RbcManager_ptr Bmc_RbcManagerAlloc(int varCapacity)
{
  return convert_to_generic_rbc_manager( Rbc_ManagerAlloc(varCapacity) );
}

void Bmc_RbcManagerReserve(RbcManager_ptr rbcManager, int newVarCapacity)
{
  Rbc_ManagerReserve( convert_to_specific_rbc_manager(rbcManager), 
		      newVarCapacity );
}

int Bmc_RbcManagerCapacity(RbcManager_ptr rbcManager)
{
  return Rbc_ManagerCapacity( convert_to_specific_rbc_manager(rbcManager) );
}

void Bmc_RbcManagerFree(RbcManager_ptr rbcManager)
{
  Rbc_ManagerFree( convert_to_specific_rbc_manager(rbcManager) );
}

void Bmc_RbcManagerGC(RbcManager_ptr rbcManager)
{
  Rbc_ManagerGC( convert_to_specific_rbc_manager(rbcManager) );
}



/**Function********************************************************************

  Synopsis           [Returns 1 if the given rbc is the true value, 
  otherwise returns 0]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_RbcIsTrue(RbcManager_ptr rbcManager, rbc_ptr arg)
{
  return (arg == Bmc_RbcTruth(rbcManager));
}

/**Function********************************************************************

  Synopsis           [Returns 1 if the given rbc is the false value, 
  otherwise returns 0]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_RbcIsFalse(RbcManager_ptr rbcManager, rbc_ptr arg)
{
  return (arg == Bmc_RbcFalsity(rbcManager));
}

/**Function********************************************************************

  Synopsis           [Builds a rbc 'true' constant value]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcTruth(RbcManager_ptr rbcManager)
{
  return convert_to_generic_rbc(
	   Rbc_GetOne( convert_to_specific_rbc_manager(rbcManager) ));
}

/**Function********************************************************************

  Synopsis           [Builds a rbc 'false' constant value]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcFalsity(RbcManager_ptr rbcManager)
{
  return convert_to_generic_rbc(
	   Rbc_GetZero( convert_to_specific_rbc_manager(rbcManager) ));
}

/**Function********************************************************************

  Synopsis           [Negates its rbc argument]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcNot(RbcManager_ptr rbcManager, rbc_ptr left)
{
   return convert_to_generic_rbc(
	   Rbc_MakeNot( convert_to_specific_rbc_manager(rbcManager), 
			convert_to_specific_rbc(left) )); 
}

/**Function********************************************************************

  Synopsis           [Builds a new rbc which is the conjunction between   
  its two arguments]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcAnd(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right)
{
  return rbc_int_makeAnd(rbcManager, left, right, RBC_TRUE);
}

/**Function********************************************************************

  Synopsis           [Builds a new rbc which is the disjunction of  
  its two arguments]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcOr(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right)
{
  return rbc_int_makeOr(rbcManager, left, right, RBC_TRUE);
}

/**Function********************************************************************

  Synopsis           [Builds a new rbc which is the exclusive-disjunction
  of its two arguments]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcXor(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right)
{
  return rbc_int_makeXor(rbcManager, left, right, RBC_TRUE);
}

/**Function********************************************************************

  Synopsis           [Builds a new rbc which is the implication between 
  its two arguments]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcImplies(RbcManager_ptr rbcManager, rbc_ptr arg1, rbc_ptr arg2)
{
  /* (a -> b) <-> !(a & !b) <-> (!a | b) */
  return rbc_int_makeOr( rbcManager, 
		     Bmc_RbcNot(rbcManager, arg1), 
		     arg2,
		     RBC_TRUE
		   );
}

/**Function********************************************************************

  Synopsis           [Builds a new rbc which is the logical equivalence 
  between its two arguments]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcIff(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right)
{
  return rbc_int_makeIff(rbcManager, left, right, RBC_TRUE);
}

/**Function********************************************************************

  Synopsis           [Builds an if-then-else operation rbc]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
rbc_ptr Bmc_RbcIte(RbcManager_ptr rbcManager, rbc_ptr c, rbc_ptr t, rbc_ptr e)
{
  return rbc_int_makeIte(rbcManager, c, t, e, RBC_TRUE);
}



/**Function********************************************************************

  Synopsis    [Creates a fresh copy G(X') of the rbc F(X) by shifting 
               each variable index of a given amount]

  Description [Shifting operation replaces each occurence of the variable x_i 
  in `f' with the variable x_(i + shift). 
  A simple lazy mechanism is implemented to optimize that cases which 
  given expression is a constant in]

  SideEffects []

  SeeAlso     []

******************************************************************************/
rbc_ptr Bmc_RbcVarShift(RbcManager_ptr rbcManager, rbc_ptr f, int shift)
{
  shift_memoize_key key;
  rbc_ptr  rbcResult = NULL;

  nusmv_assert(GL_phtRbcShift != NULL); /* init problems? */

  /* lazy evaluation: */
  if (Bmc_RbcIsTrue(rbcManager, f) || Bmc_RbcIsFalse(rbcManager, f)) {
    return f;
  }   


  /* create the key, then search for it: */
  key.rbc   = f;
  key.shift = shift;
  
  if ( !st_lookup(GL_phtRbcShift, (char*) &key, (char**) &rbcResult) ) {
    /* Duplicates the key */
    /* This ALLOC is deleted by bmcShiftHash_CallbackDeleteEntryAndKey: */
    shift_memoize_key* key_copy = ALLOC(shift_memoize_key, 1); 
    key_copy->rbc   = key.rbc;
    key_copy->shift = key.shift; 

    rbcResult = convert_to_generic_rbc(
		     Rbc_Shift( convert_to_specific_rbc_manager(rbcManager), 
		     convert_to_specific_rbc(f),
		     shift )
		    );
    
    
    st_insert( GL_phtRbcShift, (char*) key_copy, (char*) rbcResult );
  }

  nusmv_assert(rbcResult!=NULL);
  return rbcResult;
}


rbc_ptr Bmc_RbcVarSubst(RbcManager_ptr rbcManager, rbc_ptr f, int* subst)
{
   return convert_to_generic_rbc(
	   Rbc_Subst( convert_to_specific_rbc_manager(rbcManager), 
		      convert_to_specific_rbc(f),
		      subst )); 
}


rbc_ptr Bmc_RbcSubstRbc(RbcManager_ptr rbcManager, rbc_ptr f, 
			rbc_ptr* substRbc)
{
   return convert_to_generic_rbc(
	   Rbc_SubstRbc( convert_to_specific_rbc_manager(rbcManager), 
			 convert_to_specific_rbc(f),
			 convert_to_specific_rbc_ref(substRbc) )); 
}


int Bmc_RbcConvertToCnf(RbcManager_ptr rbcManager, rbc_ptr f, lsList clauses, 
			lsList vars)
{
  return Rbc_Convert2Cnf( 
			 convert_to_specific_rbc_manager(rbcManager), 
			 convert_to_specific_rbc(f),
			 clauses, 
			 vars
			);
}





void Bmc_RbcDumpDavinci(RbcManager_ptr rbcManager, rbc_ptr f, FILE* outFile)
{
  Rbc_OutputDaVinci( convert_to_specific_rbc_manager(rbcManager), 
		     convert_to_specific_rbc(f), 
		     outFile );
}

void Bmc_RbcDumpSexpr(RbcManager_ptr rbcManager, rbc_ptr f, FILE* outFile)
{
  Rbc_OutputSexpr( convert_to_specific_rbc_manager(rbcManager), 
		   convert_to_specific_rbc(f), 
		   outFile );
}



rbc_ptr Bmc_RbcGetIthVar(RbcManager_ptr rbcManager, int varIndex)
{
  return convert_to_generic_rbc(
	   Rbc_GetIthVar( convert_to_specific_rbc_manager(rbcManager), 
			  varIndex ));
}

int Bmc_RbcGetVarIndex(RbcManager_ptr rbcManager, rbc_ptr var)
{
  return Rbc_GetVarIndex( convert_to_specific_rbc(var) );
}






/*---------------------------------------------------------------------------*/
/* Definitions of internal functions                                         */
/*---------------------------------------------------------------------------*/

void Bmc_PrintStats(RbcManager_ptr rbcManager, int clustSize, FILE* outFile)
{
  Rbc_PrintStats( convert_to_specific_rbc_manager(rbcManager),
		  clustSize, 
		  outFile );
}


/**Function********************************************************************

  Synopsis           [Initializes private hast table member for shifting
  operations]

  Description        [Call BmcShiftingHashQuit() before quit from BMC module]

  SideEffects        [Private global var GL_phtRbcShift will change]

  SeeAlso            [Bmc_RbcVarShift, BmcShiftingHashQuit]

******************************************************************************/
void BmcShiftingHashInit() 
{
  /* Initializes the shifting hash table */
  nusmv_assert(GL_phtRbcShift == NULL); /* not already initialized */
 
  GL_phtRbcShift = st_init_table( &st_shift_memoize_key_cmp, 
				  &st_shift_memoize_key_hash );

  nusmv_assert(GL_phtRbcShift != NULL); /* was no able to init */
}


/**Function********************************************************************

  Synopsis           [Deletes private hast table member for shifting
  operations]

  Description        [Call BmcShiftingHashQuit() before quit from BMC module]

  SideEffects        [Private global var GL_phtRbcShift will be put to NULL]

  SeeAlso            [Bmc_RbcVarShift, BmcShiftingHashInit]

******************************************************************************/
void BmcShiftingHashQuit() 
{
  /* Destroys the shifting hash table */
  if (GL_phtRbcShift != NULL) {
    st_foreach(GL_phtRbcShift, bmcShiftHash_CallbackDeleteEntryAndKey, 
	       NULL /*dummy*/);

    st_free_table(GL_phtRbcShift);
    GL_phtRbcShift = NULL;
  }
}


/*---------------------------------------------------------------------------*/
/* Definitions of static functions                                           */
/*---------------------------------------------------------------------------*/

static rbc_ptr 
rbc_int_makeAnd(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right,
		rbc_bool_t sign)
{
   return convert_to_generic_rbc(
	   Rbc_MakeAnd( convert_to_specific_rbc_manager(rbcManager), 
			convert_to_specific_rbc(left),
			convert_to_specific_rbc(right), 
			convert_to_specific_rbc_bool(sign) )); 
}


static rbc_ptr 
rbc_int_makeOr(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right, 
	       rbc_bool_t sign)
{
   return convert_to_generic_rbc(
	   Rbc_MakeOr( convert_to_specific_rbc_manager(rbcManager), 
		       convert_to_specific_rbc(left),
		       convert_to_specific_rbc(right), 
		       convert_to_specific_rbc_bool(sign) )); 
}

static rbc_ptr 
rbc_int_makeXor(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right,
		rbc_bool_t sign)
{
   return convert_to_generic_rbc(
	   Rbc_MakeXor( convert_to_specific_rbc_manager(rbcManager), 
		        convert_to_specific_rbc(left),
		        convert_to_specific_rbc(right), 
		        convert_to_specific_rbc_bool(sign) )); 
}

static rbc_ptr 
rbc_int_makeIff(RbcManager_ptr rbcManager, rbc_ptr left, rbc_ptr right,
		rbc_bool_t sign)
{
   return convert_to_generic_rbc(
	   Rbc_MakeIff( convert_to_specific_rbc_manager(rbcManager), 
			convert_to_specific_rbc(left),
			convert_to_specific_rbc(right), 
			convert_to_specific_rbc_bool(sign) )); 
}

static rbc_ptr 
rbc_int_makeIte(RbcManager_ptr rbcManager, rbc_ptr c, rbc_ptr t, 
		rbc_ptr e, rbc_bool_t sign)
{
   return convert_to_generic_rbc(
	   Rbc_MakeIte( convert_to_specific_rbc_manager(rbcManager), 
			convert_to_specific_rbc(c),
			convert_to_specific_rbc(t),
			convert_to_specific_rbc(e),
			convert_to_specific_rbc_bool(sign) )); 
}

static rbc_ptr rbc_getLeftOpnd(rbc_ptr f)
{
  return convert_to_generic_rbc( Rbc_GetLeftOpnd(convert_to_specific_rbc(f)) );
}

static rbc_ptr rbc_getRightOpnd(rbc_ptr f)
{
  return convert_to_generic_rbc(Rbc_GetRightOpnd(convert_to_specific_rbc(f)));
}


static void rbc_mark(RbcManager_ptr rbc, rbc_ptr f)
{
  Rbc_Mark( convert_to_specific_rbc_manager(rbc), 
	    convert_to_specific_rbc(f) );
}

static void rbc_unmark(RbcManager_ptr rbc, rbc_ptr f)
{
  Rbc_Unmark( convert_to_specific_rbc_manager(rbc), 
	      convert_to_specific_rbc(f) );
}


/*---------------------------------------------------------------------------*/
/* Gateway functions: implementation changes depending on the specific rbc   */
/* package is being used.                                                    */

static rbc_ptr  convert_to_generic_rbc(RBC_SPEC_TYPE_P pRbc) 
{ 
  return (rbc_ptr) pRbc; 
}

static RBC_SPEC_TYPE_P convert_to_specific_rbc(rbc_ptr pRbc) 
{
  return (RBC_SPEC_TYPE_P) pRbc;
} 

static rbc_ptr* convert_to_generic_rbc_ref(RBC_SPEC_TYPE_P* pRbcRef)
{
  return (rbc_ptr*) pRbcRef;
}

static RBC_SPEC_TYPE_P* convert_to_specific_rbc_ref(rbc_ptr* pRbcRef)
{
  return (RBC_SPEC_TYPE_P*) pRbcRef; 
}

static RBC_SPEC_MANAGER_TYPE_P 
convert_to_specific_rbc_manager(RbcManager_ptr pRbcManager)
{
  return (RBC_SPEC_MANAGER_TYPE_P) pRbcManager; 
}

static RbcManager_ptr 
convert_to_generic_rbc_manager(RBC_SPEC_MANAGER_TYPE_P pRbcManager)
{
  return (RbcManager_ptr) pRbcManager;
}

static RBC_SPEC_BOOL_TYPE convert_to_specific_rbc_bool(rbc_bool_t rbcBool)
{
  return (RBC_SPEC_BOOL_TYPE) rbcBool;
}

static rbc_bool_t  convert_to_generic_rbc_bool(RBC_SPEC_BOOL_TYPE rbcBool)
{
  return (rbc_bool_t) rbcBool;
}



/* ------------------------------------------------------------------------  */
/* Shift memoizing internal functions:                                       */
 
static int st_shift_memoize_key_cmp(const char* _key1, const char* _key2) 
{
  const shift_memoize_key* key1 = (shift_memoize_key*) _key1;
  const shift_memoize_key* key2 = (shift_memoize_key*) _key2;
  
  return (key1->rbc != key2->rbc) | (key1->shift != key2->shift);
}

static int st_shift_memoize_key_hash(char* _key, const int size) 
{
  shift_memoize_key* key = (shift_memoize_key*) _key;
  return ((unsigned int)((int)key->rbc >> 2) ^ (key->shift) ) % size;
}  


static enum st_retval 
bmcShiftHash_CallbackDeleteEntryAndKey(char* key, char* record, char* dummy)
{
  FREE(key); /* removes allocated key for this entry */
  return ST_DELETE; /* removes associated element */
}
    
   






