/**CFile***********************************************************************

  FileName    [bmcTest.c]

  PackageName [bmc]

  Synopsis    [Test routines for <tt>bmc</tt> package]

  Description []

  SeeAlso     []

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``bmc'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst and University of Trento. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define GEN_WFF_CONSES_OP_NUMBER 10

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
static int iGeneratedFormulae = 0;

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static node_ptr bmc_gen_wff ARGS((int max_depth, int max_conns));

static node_ptr bmc_gen_tableau_test ARGS((node_ptr ltl_nnf_wff, 
					   int k, int l));

static int UsageBmcTestTableau ARGS((void));
static void BmcBexpOutput ARGS((FILE* f, node_ptr bexp, int output_type));


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Call this function to reset the test sub-package (into 
  the reset command for example)]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void Bmc_TestReset()
{
  iGeneratedFormulae = 0;
}





/**Function********************************************************************

  Synopsis           [For each variable p in the set of state variables, 
  generates the global equivalence of p and X^(loop length), starting from
  the loop start]

  Description [ In the following example we suppose the loop starts
  from 2 and finishes to 5 (the bound).

  <PRE>
        ,--------,
        V        |       
  o--o--o--o--o--o--o--o--o--o--o--o--o- (...continues indefinitely)
  0  1  2  3  4  5  6  7  8  9  10 11 12 

  </PRE>


  In general all variables in time 2 must be forced to be equivalent
  to the corresponding variables timed in 6, the variables in 3 to 7,
  and so on up to the variables in 5 (equivalent to variables in
  9). Then variables in 6 (or 2 again) must be forced to be equivalent
  to the varaibles in 10, and so on indefinitely.
  <BR><BR>
  In formula (let suppose we have only one boolean variable):
  <BR>
  (p2 <-> p6) && (p6 <-> p10) ...
  <BR><BR>
  In a more compact (and finite!) form, related to this example:
  XX(G (p <-> XXXX(p)))

  The first two neXtes force the formula to be effective only from the loop
  starting point. 
  The generic formula implemented in the code is the following one:
  <PRE>
  X^(l) (G ( 
             (p0 <-> X^(k-l+1)(p0)) && 
	     (p1 <-> X^(k-l+1)(p1)) &&
	                .
		        .
		        .
	     (pn <-> X^(k-l+1)(pn))
	   )
        )
  </PRE>
 where:
   p0..pn are all boolean variables into the model
   X^(n) is expanded to XXX..X n-times ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr Bmc_MkLoopbackLtl(int k, int l)
{
  node_ptr result;
  node_ptr bigand_vars;
  node_ptr single_var_eq;
  node_ptr var;
  int i = 0;
  int iLoopLength = 0;

  nusmv_assert( (l != BMC_NO_LOOP) && (l <= k) );
  nusmv_assert( Bmc_VarsMgrGetNum() > 0 );

  iLoopLength = k-l+1;
  
  /* first cycle is performed manually, in order to optimize a bit */
  var = Bmc_GetIthStatevar(i);
  bigand_vars = Bmc_WffMkIff( var, Bmc_WffMkXopNext(var, iLoopLength) );  

  /* iterates across the remaining variables: */
  for (i = 1; i < Bmc_VarsMgrGetNum(); i++) {
    var = Bmc_GetIthStatevar(i);
    single_var_eq = Bmc_WffMkIff( var, Bmc_WffMkXopNext(var, iLoopLength) );
    bigand_vars = Bmc_WffMkAnd(bigand_vars, single_var_eq);
  }

  result = Bmc_WffMkGlobally(bigand_vars);
  result = Bmc_WffMkXopNext(result, l); /* shifts to loop starting point */

  return result;
}




/**Function********************************************************************

  Synopsis           [The first time Bmc_TestTableau is called in the current
  session this function creates a smv file with a model and generates a random
  ltl spec to test tableau. The following times it is called it appends a new
  formula to the file.]

  Description        [If you call this command with a loopback set to 
  BMC_ALL_LOOPS you command execution is aborted.]

  SideEffects        []

  SeeAlso            []

  CommandName        [_bmc_test_tableau] 	   

  CommandSynopsis    [Generates a random formula to logically test the 
  equivalent tableau]  

  CommandArguments   [\[-h\] | \[-n property_index\] | \[\[ -d max_depth\] \[-c max_conns\] \[-o operator\]\]
  ]  

  CommandDescription [Use this hidden command to generate random formulae and 
  to test the equivalent tableau. The first time this command is called in the
  current NuSMV session it creates a new smv file with a model and generates a
  random ltl spec to test tableau. 
  The following times it is called it appends a new formula to the file. 
  The generated model contains the same number of non-deterministic variables
  the currently model loaded into NuSMV contains. <BR>
  You cannot call this command if the bmc_loopback is set to '*' (all loops).
  ]  

******************************************************************************/
int Bmc_TestTableau(int argc, char ** argv) 
{
  node_ptr tableau_test;
  node_ptr wff=NULL;
  int k,l,max_depth, max_conns;
  char szLoopback[16];
  FILE * f;
  /* user can generate a random wff based on a specified operator: */
  enum GenWffOperator { 
    GWO_None, GWO_Globally, GWO_Future, GWO_Until, GWO_Releases  
  } wff_operator;

  nusmv_assert(iGeneratedFormulae>=0);

  wff_operator = GWO_None;
  
  k = get_bmc_pb_length(options); 
  l = get_bmc_pb_loop(options);

  if (l == BMC_ALL_LOOPS) {
    /* not implemented yet */
    fprintf (nusmv_stderr, "Error: the case 'all loops' is not implemented yet.\nPlease set the variable 'bmc_loopback' to another value.\n\n");
    return(1);
  }

  max_depth = -1;  max_conns = 1; /* default values */

  if (cmp_struct_get_bmc_setup(cmps) == 0) {
    fprintf (nusmv_stderr, "Please call bmc_setup before use this command.\n");
    return(1);
  }

  /* process command options */ 
  {
    int c;
    char* strNumber = NIL(char);
    char* szOperator= NIL(char);
    int prop_no;
    Prop_Ptr ltlprop=(Prop_Ptr)NULL;   /* The property being processed */
    node_ptr ltlspec;


    
    util_getopt_reset();
    while((c = util_getopt(argc, argv, "hn:d:c:o:")) != EOF) {
      switch (c) {
      case 'h':
	return UsageBmcTestTableau();
      case 'n':
	{
	  strNumber = util_strsav(util_optarg);
	  {
	    char* err_occ[2];

	    err_occ[0] = "";
	    prop_no = strtol(strNumber, err_occ, 10);
	    if ((strcmp(err_occ[0], "") != 0)) {
	      fprintf(nusmv_stdout,
		      "Error: \"%s\" is not a valid value (must be integer).\n"
		      , strNumber);
	      return(1);
	    }
	  }
	  if (prop_no >= Prop_Db_GetSize() || prop_no < 0) {
	    fprintf(nusmv_stdout,
		    "Error: \"%s\" is not a valid value, must be in the range [0,%d].\n", 
		    strNumber, Prop_Db_GetSize()-1);
	    return(1);
	  }

	  ltlprop = Prop_Db_GetNum(prop_no);

	  if (Prop_GetType(ltlprop) != Prop_Ltl) {
	    fprintf(nusmv_stderr,
		    "Error: property %d is not of type LTL\n", prop_no);
	    return(1);
	  }
	  
	  /* here prop is ok */
	  ltlspec = Prop_GetProp(ltlprop);
	  ltlspec = Compile_FlattenSexpExpandDefine(ltlspec, Nil);
	  wff = Bmc_LtlWffMkNnfToplevel(detexpr2bexpr(ltlspec));  
	  break;
	}

      case 'd': /* for max_depth */
	{
	  strNumber = util_strsav(util_optarg);
	  {
	    char* err_occ[2];

	    err_occ[0] = "";
	    max_depth = strtol(strNumber, err_occ, 10);
	    if ((strcmp(err_occ[0], "") != 0)) {
	      fprintf(nusmv_stdout,
		    "Error: \"%s\" is not a valid value (must be integer).\n",
		     strNumber);
	      return(1);
	    }
	  }
	}
	break;

      case 'c': /* for max_conns */
	{
	  strNumber = util_strsav(util_optarg);
	  {
	    char* err_occ[2];

	    err_occ[0] = "";
	    max_conns = strtol(strNumber, err_occ, 10);
	    if ((strcmp(err_occ[0], "") != 0)) {
	      fprintf(nusmv_stdout,
		    "Error: \"%s\" is not a valid value (must be integer).\n",
		     strNumber);
	      return(1);
	    }
	  }
	}
	break;

      case 'o': /* operator specification */
	{
	  szOperator = util_strsav(util_optarg);
	  if (strcmp(szOperator, "G")==0) wff_operator = GWO_Globally;
	  else if (strcmp(szOperator, "F")==0) wff_operator = GWO_Future;
	  else if (strcmp(szOperator, "U")==0) wff_operator = GWO_Until;
	  else if (strcmp(szOperator, "R")==0) wff_operator = GWO_Releases;
	  
	  if(wff_operator == GWO_None) {
	    fprintf(nusmv_stdout,
		    "Error: operator \"%s\" is not valid. Use G|F|X|U|R\n", 
		    szOperator);
	    return(1);    
	  }
	  break;
	}
	    
      } /* switch */
      
    } /* while */
    
    if ((argc != 3) && (argc != 5) && (argc != 7)) {
      return(UsageBmcTestTableau());
    }
  
  }


  if (wff == NULL) {
    /* generates a random wff: */
    switch(wff_operator) {
    case GWO_None:
      wff = bmc_gen_wff(max_depth, max_conns);
      break;

    case GWO_Globally:
      wff = Bmc_WffMkGlobally(bmc_gen_wff(max_depth, max_conns));
      break;

    case GWO_Future:
      wff = Bmc_WffMkEventually(bmc_gen_wff(max_depth, max_conns));
      break;

    case GWO_Until:
      wff = Bmc_WffMkUntil(bmc_gen_wff(max_depth, max_conns), 
			   bmc_gen_wff(max_depth, max_conns));
      break;

    case GWO_Releases:
      wff = Bmc_WffMkReleases(bmc_gen_wff(max_depth, max_conns), 
			      bmc_gen_wff(max_depth, max_conns));
      break;
      
    default:
      nusmv_assert(FALSE); /* no other types are expected here */
    }
    wff = Bmc_LtlWffMkNnfToplevel(wff);
  }

  /* generates the test tableau */
  tableau_test = bmc_gen_tableau_test(wff, k,l);

  /* writes down the imply formula */
  if (iGeneratedFormulae == 0) {
    int i=0;

    f = fopen("Bmc_TestTableau.smv", "w");
    nusmv_assert(f != NULL);

    /* writes down the non-deterministic model */
    fprintf(f, "MODULE main\nVAR\n");
    for (i = 0; i < Bmc_VarsMgrGetNum(); i++) {
      fprintf(f, "p%d: boolean;\n", i);
    }
  }
  else {
    /* this command has been already invoked */
    f = fopen("Bmc_TestTableau.smv", "a");
    nusmv_assert(f != NULL);
  }
  
  Bmc_ConvertLoopbackFromInteger(l, szLoopback, sizeof(szLoopback));
  fprintf(f, "\n\n-- Property %d (k=%d, l=%s, max_depth=%d, max_conns=%d): \n",
	  iGeneratedFormulae, k, szLoopback, max_depth, max_conns);
  fprintf(f, "LTLSPEC ");

  ++iGeneratedFormulae;  

  fprintf (f, "\n");
  BmcBexpOutput(f, tableau_test, BMC_BEXP_OUTPUT_SMV);
  fprintf(f, "\n\n");  

  fclose(f);
  return 0;
}




/**Function********************************************************************

  Synopsis           [Usage string for Bmc_TestTableau]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static int UsageBmcTestTableau(void)
{
  fprintf (nusmv_stderr, 
"usage: _bmc_test_tableau [-h] | [-n <property_index>] | \n \
                         [[ -d <max_depth>] [-c <max_conns>] [-o <operator>]]\n");
  fprintf (nusmv_stderr, "  -h \t\t\tPrints the command usage.\n");
  fprintf (nusmv_stderr, "  -n <prop_idx> \tTest tableau of property represented by the given index\n"); 
  fprintf (nusmv_stderr, "  -d <max_depth>\tGenerates a random wff with the given max depth (default value is -1)\n");
  fprintf (nusmv_stderr, "  -c <max_conns>\tGenerates a random wff with the given max number of connectives\n\t\t\t(default value is 1).\n");
  fprintf (nusmv_stderr, "  -o <operator> \tGenerates a random wff based on the specified operator, which will \n\t\t\tbe placed at top level. Valid values are G | F | U | R\n");
  return(1);
}




/**Function********************************************************************

  Synopsis           [Bmc_TestVarsMgr]

  Description        []

  SideEffects        [...]

  SeeAlso            []

******************************************************************************/
void Bmc_TestVarsMgr()
{
  int i;
  node_ptr X, X1, X2, X3, X4;
  rbc_ptr A, A1, A2, A3, A4,
    B, B1, B2, B3, B4,
    C, C1, C2, C3, C4;
  boolean local_test_passed, test_passed = true;
  int time;

  printf("maxtime = %d\n", Bmc_VarsMgrGetMaxtime());

  for (i = 0; i < Bmc_VarsMgrGetNum(); i++) {
    for (time = 0; time <= Bmc_VarsMgrGetMaxtime(); time++) {
      X = Bmc_GetIthStatevar(i);
      A = Rbc_GetIthVar(Bmc_VarsMgrGetRbcMgr(), i);
      B = Rbc_GetIthVar( Bmc_VarsMgrGetRbcMgr(), i + Bmc_VarsMgrGetNum() );
      C = Rbc_GetIthVar( Bmc_VarsMgrGetRbcMgr(),
			  i + ((2 + time) * Bmc_VarsMgrGetNum()) );

      X1 = Bmc_GetIthStatevar(i);
      X2 = Bmc_Curvar2Statevar(A);
      X3 = Bmc_Nextvar2Statevar(B);
      X4 = Bmc_Timedvar2Statevar(C);

      A1 = Bmc_GetIthCurvar(i);
      A2 = Bmc_Statevar2Curvar(X);
      A3 = Bmc_Nextvar2Curvar(B);
      A4 = Bmc_Timedvar2Curvar(C);

      B1 = Bmc_GetIthNextvar(i);
      B2 = Bmc_Statevar2Nextvar(X);
      B3 = Bmc_Curvar2Nextvar(A);
      B4 = Bmc_Timedvar2Nextvar(C);

      C1 = Bmc_GetIthTimedvar(i, time);
      C2 = Bmc_Statevar2Timedvar(X, time);
      C3 = Bmc_Nextvar2Timedvar(B, time);
      C4 = Bmc_Curvar2Timedvar(A, time);

      local_test_passed = (A1 == A2) && (A2 == A3) && (A3 == A4) &&
	(B1 == B2) && (B2 == B3) && (B3 == B4) &&
	(C1 == C2) && (C2 == C3) && (C3 == C4);

      printf("X = %p\tX1 = %p\tX2 = %p\tX3 = %p\tX4 = %p\n", X, X1, X2,
	      X3, X4);
      printf("A = %p\tA1 = %p\tA2 = %p\tA3 = %p\tA4 = %p\n", A, A1, A2,
	      A3, A4);
      printf("B = %p\tB1 = %p\tB2 = %p\tB3 = %p\tB4 = %p\n", B, B1, B2,
	      B3, B4);
      printf("C = %p\tC1 = %p\tC2 = %p\tC3 = %p\tC4 = %p\n", C, C1, C2,
	      C3, C4);
      printf("%d-ith variable, @%d time\n", i, time);
      printf("Test [%s]\n", local_test_passed ? "PASSED" : "NOT PASSED");

      test_passed = test_passed && local_test_passed;
    }
  }

  printf("Number of variables = %d\n", Bmc_VarsMgrGetNum());
  printf("X = curvar_bexpr\nA = curvar\nB = nextvar\nC = timedvar\n");

  printf("\n\nWP4_test1b [%s]\n", test_passed ? "PASSED" : "\aNOT_PASSED");
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Given a WFF in NNF, converts it into a tableau 
  formula, then back to WFF_(k,l) and returns WFF -> WFF_(k,l)]

  Description        [This function is used to test tableau formulae]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static node_ptr bmc_gen_tableau_test(node_ptr ltl_nnf_wff, int k, int l)
{
  node_ptr tableau_as_wff;
  node_ptr implies_formula; 
  rbc_ptr tableau_k_l_ltl_wff;

  /* generates the tableau: */
  tableau_k_l_ltl_wff = Bmc_MkTableau_k_l(ltl_nnf_wff, k, l);

  /* reconvert the tableau back to a wff_(k,l) */
  tableau_as_wff = Bmc_Rbc2Bexp(tableau_k_l_ltl_wff);

  /* build the implies: */
  if (l == BMC_NO_LOOP) {
    implies_formula = Bmc_WffMkImplies(tableau_as_wff, ltl_nnf_wff);
  }
  else {
    nusmv_assert(l != BMC_ALL_LOOPS); /* all loops are not allowed nowadays */
    implies_formula = 
      Bmc_WffMkImplies(Bmc_WffMkAnd(tableau_as_wff, Bmc_MkLoopbackLtl(k, l)),
		       ltl_nnf_wff 
		      );
  }

  return implies_formula;
}


/**Function********************************************************************

  Synopsis           [Builds a <b>random LTL WFF</b> with specified 
  <tt>max</tt> depth and <tt>max</tt> connectives.]

  Description        []

  SideEffects        [node hash may change]

  SeeAlso            []

******************************************************************************/
static node_ptr bmc_gen_wff(int max_depth, int max_conns)
{
  int rnd;

  /* generate a random number (rnd >= 1) &&
     (rnd <= (GEN_WFF_CONSES_OP_NUMBER + 'number of state variables')) */
  rnd = 1 + (int) ((float) (GEN_WFF_CONSES_OP_NUMBER +
		   Bmc_VarsMgrGetNum()) * rand() /
		   (RAND_MAX + 1.0));

  /* if depth or connses of wff are exausted get a random number such that:
     (rnd >= 0) && (rnd < 'number of state variables')... */
  if ((max_depth < 0) || (max_conns < 0)) {
    rnd = (int) ((float) Bmc_VarsMgrGetNum() * rand() /
		 (RAND_MAX + 1.0));
    /* ...then return correspondent state variable to the random integer */
    return Bmc_GetIthStatevar(rnd);
  }

  /* exclude atoms from depth and connses decrement contibutes */
  if (rnd <= GEN_WFF_CONSES_OP_NUMBER) {
    --max_depth; --max_conns;
  }

  switch (rnd) {
  case 1:
    return Bmc_WffMkNot(bmc_gen_wff(max_depth, max_conns));
  case 2:
    return Bmc_WffMkEventually(bmc_gen_wff(max_depth, max_conns));
  case 3:
    return Bmc_WffMkGlobally(bmc_gen_wff(max_depth, max_conns));
  case 4:
    return Bmc_WffMkOpNext(bmc_gen_wff(max_depth, max_conns));
  case 5:
    return Bmc_WffMkImplies(bmc_gen_wff(max_depth, max_conns),
			   bmc_gen_wff(max_depth, max_conns));
  case 6:
    return Bmc_WffMkIff(bmc_gen_wff(max_depth, max_conns),
		       bmc_gen_wff(max_depth, max_conns));
  case 7:
    return Bmc_WffMkOr(bmc_gen_wff(max_depth, max_conns),
		      bmc_gen_wff(max_depth, max_conns));
  case 8:
    return Bmc_WffMkAnd(bmc_gen_wff(max_depth, max_conns),
		       bmc_gen_wff(max_depth, max_conns));
  case 9:
    return Bmc_WffMkUntil(bmc_gen_wff(max_depth, max_conns),
			 bmc_gen_wff(max_depth, max_conns));

  case 10:
   return Bmc_WffMkReleases(bmc_gen_wff(max_depth, max_conns),
         bmc_gen_wff(max_depth, max_conns)); 

  default:
    return Bmc_GetIthStatevar(rnd - GEN_WFF_CONSES_OP_NUMBER - 1);
  }
}




/**Function********************************************************************

  Synopsis    [<b>Write</b> to specified FILE stream given node_ptr
  <b>formula</b> with specified <tt>output_type</tt> format. There are
  follow formats: <tt>BMC_BEXP_OUTPUT_SMV, BMC_BEXP_OUTPUT_LB</tt>]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static void BmcBexpOutput(FILE* f, node_ptr bexp, int output_type)
{
  int type;

  nusmv_assert(f != NULL);

  /* exit from recursion if given formula is Nil */
  if (bexp == Nil) return;

  /* assert that input formula type can't be a NEXT operator, that is
     used in model specification (section ASSIGN). We use here only OP_NEXT
     operator used in the module (section LTLSPEC). */
  nusmv_assert (node_get_type (bexp) != NEXT);

  type = node_get_type (bexp);

  switch (type) {
  case FALSEEXP:                        /* FALSEEXP  */
    fprintf (f, "%s", (output_type == BMC_BEXP_OUTPUT_SMV) ? "0" : "false");
    break;

  case TRUEEXP:                         /* TRUEEXP   */
    fprintf (f, "%s", (output_type == BMC_BEXP_OUTPUT_SMV) ? "1" : "true");
    break;

  case AND:                             /* AND       */
    fprintf(f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf(f, " %s ", (output_type == BMC_BEXP_OUTPUT_SMV) ? "&" : "/\\");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf(f, ")");
    break;

  case OR:                              /* OR        */
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, " %s ", (output_type == BMC_BEXP_OUTPUT_SMV) ? "|" : "\\/");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf (f, ")");
    break;

  case NOT:                             /* NOT       */
    fprintf (f, "%c", (output_type == BMC_BEXP_OUTPUT_SMV) ? '!' : '~');
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, ")");
    break;

  case IMPLIES:                         /* IMPLIES   */
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, " -> ");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf (f, ")");
    break;

  case IFF:                             /* IFF       */
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, " <-> ");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf (f, ")");
    break;

  case OP_FUTURE:                       /* OP_FUTURE */
    fprintf (f, "F(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, ")");
    break;

  case OP_GLOBAL:                       /* OP_GLOBAL */
    fprintf (f, "G(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, ")");
    break;

  case UNTIL:                           /* UNTIL     */
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, " U ");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf (f, ")");
    break;

  case RELEASES:                        /* RELEASES  */
    fprintf (f, "(");
    BmcBexpOutput(f, car (bexp), output_type);
    fprintf (f, " V ");
    BmcBexpOutput(f, cdr (bexp), output_type);
    fprintf (f, ")");
    break;

  case OP_NEXT:                         /* OP_NEXT   */
    {
      int i = 0;
      {
	/* prints out "X(" suffix while OP_NEXT is encountred */
	do {
	  fprintf (f, "X("); 
	  bexp = car (bexp);
	  nusmv_assert(bexp != Nil);
	  i++;
	} while (node_get_type (bexp) == OP_NEXT);
      }

      /* then print the internal bexp */
      BmcBexpOutput(f, bexp, output_type);

      while ((i--) > 0) fprintf (f, ")");
    }
    break;

  default:                              /* (default action) */
    {
      rbc_ptr r;

      /* gets the the rbc correspondent to the state variable */
      r = Bmc_Statevar2Curvar (bexp);

      /* if bexp is a really state variable, then prints out the index
	 of correspondent rbc variable */
      if (r != (rbc_ptr)NULL) {
	fprintf(f, "p%d", Rbc_GetVarIndex (r));
      }
      else {
	internal_error("BmcBexpOutput: given wff atom isn\' in RBC environ\n");
      }
    }
  }

}





