/**CFile***********************************************************************

  FileName    [compileAffinity.c]

  PackageName [compile]

  Synopsis    [Affinity clustering optimized.]

  Description [Affinity clustering optimized.]

  SeeAlso     []

  Author      [Emanuele Olivetti]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2. 
  Copyright (C) 2000-2001 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/


#include "compileInt.h" 
#include "heap.h"

static char rcsid[] UTIL_UNUSED = "$Id: compileAffinity.c,v 1.7 2001/11/05 10:57:01 none Exp $";

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/**Struct**********************************************************************

  Synopsis    [record of the list to be used with the heap]

  Description [This is the record of the list to be used with the heap, basic
  blocks of optimized affinity clustering.]

  SeeAlso     []

******************************************************************************/
struct list_entry_ {
  boolean  flag; /* tells if the cluster exists */
  CPCluster_Info_t * C;    /* pointer to the element (cluster) of the original cluster list */
};

typedef struct list_entry_ list_entry;

/**Struct**********************************************************************

  Synopsis    [pair of pointers to the list.]

  Description [This is the record of the heap that point to the support list.]

  SeeAlso     []

******************************************************************************/
/* contiene la coppia di puntatori a cluster presenti nella lista */
struct pair_ {
  list_entry * C1;
  list_entry * C2;
};

typedef struct pair_ pair;


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Allocs a list_entry]

  Description        [Allocs a list_entry]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static list_entry * list_entryAlloc(void)
{
  list_entry * L_E;

  L_E = ALLOC(list_entry, 1);
  L_E->flag        = false;
  L_E->C           = NULL;
  return(L_E);
}

/**Function********************************************************************

  Synopsis           [Allocates a pair]

  Description        [Allocates a pair]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static pair * pairAlloc(void)
{
  pair * P;

  P = ALLOC(pair, 1);
  P->C1          = NULL;
  P->C2          = NULL;
  return(P);
}


/**Function********************************************************************

  Synopsis           [Compute the Affinity of two BDD clusters.]

  Description        [Compute the Affinity between two BDD clusters as 
  suggested by Moon, Hachtel, Somenzi in BBT paper. Affinity is the ratio
  between the number of shared variables and the number of the union of
  all variables (intersection/union)]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static double bdd_compute_affinity(bdd_ptr a, bdd_ptr b)
{
  double result;
  bdd_ptr supp_a, supp_b, I , U;
  supp_a = bdd_support(dd_manager,a);
  supp_b = bdd_support(dd_manager,b);
  I = bdd_cube_intersection(dd_manager, supp_a, supp_b);
  U = bdd_cube_union(dd_manager, supp_a, supp_b);
  bdd_free(dd_manager, supp_a);
  bdd_free(dd_manager, supp_b);
  if (bdd_size(dd_manager,U)!=0){
    result = ((double)bdd_size(dd_manager,I))/((double)bdd_size(dd_manager,U));
  }else{
    /* a,b are empty cubes!*/
    abort();
  }
  bdd_free(dd_manager, I);
  bdd_free(dd_manager, U);

  return(result);
}


/**Function********************************************************************

  Synopsis           [Add a new entry in support list and new pairs in heap.]

  Description        [Add a new entry in support list and new pairs in heap.
  Pairs with a dead cluster are skipped]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static node_ptr list_heap_add(CPCluster_Info_t * C, node_ptr L, heap H)
{
  node_ptr List=L;
  list_entry * New_entry;
  bdd_ptr T1;

  New_entry = list_entryAlloc();
  New_entry->flag = true;
  New_entry->C = C;
  
  T1 = ((CPCluster_Info_t *)C)->Ti;
  while(List != Nil){
    if (((list_entry*)(car(List)))->flag == true){
      pair * P = pairAlloc();
      bdd_ptr T2=((list_entry*)car(List))->C->Ti;
      double affinity=bdd_compute_affinity(T1,T2);
      P->C1=New_entry;
      P->C2=(list_entry *)car(List);
      heap_add(H, affinity, P);
    }
    List=cdr(List);
  }

  L = cons((node_ptr)New_entry,L);

  return(L);
}

/**Function********************************************************************

  Synopsis           [Delete a cluster in support list.]

  Description        [Delete a cluster in support list.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static void list_del(list_entry * C)
{
  C->flag=false;
  return;
}

/**Function********************************************************************

  Synopsis           [Copy over threshold clusters in result list or in support
  list & heap.]

  Description        [Copy over threshold clusters in result list or in support
  list & heap. It doesn't modify the input list.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static int affinity_move_clusters(node_ptr List, node_ptr *NewRelationList, int threshold, node_ptr * L, heap H)
{
  int n=0;
  while (List != Nil) {
    CPCluster_Info_t * Ci = (CPCluster_Info_t *)car(List);
    bdd_ptr Ti = bdd_dup(Ci->Ti);
    if (bdd_size(dd_manager,Ti)>threshold){
      CPCluster_Info_t * New_Ci = CPClusterInfoAlloc();
      New_Ci->Ti = Ti;
      *NewRelationList = cons((node_ptr)New_Ci, *NewRelationList);
    } else {
      *L=list_heap_add((CPCluster_Info_t *)car(List), *L, H);
      n++;
    }
    List = cdr(List);
  }
  return(n);
}

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [OPTIMIZED affinity clustering]

  Description        [OPTIMIZED affinity clustering]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void compileCPMakeAffinityClusters(DdManager * dd, node_ptr RelationList,
				   node_ptr * NewRelationList,
				   int threshold)
{
  node_ptr L=Nil;
  heap H = heap_create();
  pair * P;
  int n;
  node_ptr oldList;

  CPDupClustersList(dd, RelationList, &oldList);

  *NewRelationList = Nil;

  n = affinity_move_clusters(oldList, NewRelationList, threshold, &L, H);

  /* clustering */
  while (n>1) {
    P = (pair*)heap_getmax(H);
      if (P->C1->flag==true && P->C2->flag==true) {
	bdd_ptr T1 , T2 , New_T;
	CPCluster_Info_t * New_C;
	T1=P->C1->C->Ti;
	T2=P->C2->C->Ti;
	New_T = bdd_and(dd, T1, T2);
	list_del(P->C1);
	list_del(P->C2);
	
	New_C = CPClusterInfoAlloc();
	New_C->Ti = New_T;
	
	if (bdd_size(dd,New_T)>threshold) {
	  *NewRelationList=cons((node_ptr)New_C, *NewRelationList);
	  n-=2;
	} else {
	  L=list_heap_add(New_C, L, H);
	  n--;
	}
      }
  }
  
  /* there can be a last one small cluster */
  if (n==1) {
    node_ptr tmplist = L;
    while((tmplist!=Nil) && (((list_entry*)(car(tmplist)))->flag == false)){
      tmplist=cdr(tmplist);
    }
    if (tmplist!=Nil){
      CPCluster_Info_t * C;
      bdd_ptr T;
      C = CPClusterInfoAlloc();
      T=((list_entry*)car(tmplist))->C->Ti;
      C->Ti=bdd_dup(T);
      *NewRelationList=cons((node_ptr)C, *NewRelationList);
    } else {
      fprintf(nusmv_stdout,"Affinity Optimized Inconsitency!!!\n");
      abort();
    }
  }

  /* collecting garbage... */
  {
    node_ptr tmplist = L;
    while(tmplist!=Nil){
      bdd_ptr T;
      T=((list_entry*)car(tmplist))->C->Ti;
      bdd_free(dd,T);
      tmplist=cdr(tmplist);
    }
    free_list(L);
    while(heap_isempty(H)==0){
      P=heap_getmax(H);
    }
    heap_destroy(H);
  }
}
